--
-- TreeShear
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history:	v1.0 - 2016-04-10 - Initial implementation
-- 

TreeShear = {};

TreeShear.STATE_READY = 0;
TreeShear.STATE_CUTTING_TREE = 1;
TreeShear.STATE_CUTTING_TREE_IS_FALLING = 2;
TreeShear.STATE_RETURN = 3;

TreeShear.CHANGE_RESET = 0;
TreeShear.CHANGE_CUT = 1;
TreeShear.CHANGE_FALLING = 2;
TreeShear.CHANGE_RETURN = 3;

function TreeShear.prerequisitesPresent(specializations)
	if not SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations) then print("Warning: Specialization TreeShear needs the specialization AnimatedVehicle."); end;
	
	return SpecializationUtil.hasSpecialization(AnimatedVehicle, specializations);
end;

function TreeShear:load(xmlFile)
	self.setShearState = SpecializationUtil.callSpecializationsFunction("setShearState");
	self.setHeadState = SpecializationUtil.callSpecializationsFunction("setHeadState");
	
	local inRangeNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.shearAnimation#playerInRangeNode"));
	if inRangeNode ~= nil then 
		self.playerInRangeTool = {};
		self.playerInRangeTool[1] = {};
		self.playerInRangeTool[1].node = inRangeNode;
		self.playerInRangeTool[1].playerDistance = 2;
		self.playerInRangeTool[1].interactiveData = {text1 = g_i18n:getText("ROTATE_SHEAR"), functionName = "setShearState", boolName = "shearState", offset = {0, 0, 0}};
	end;
	
	self.lookForSplitShape = self.attachHydraulicCoupling == nil;
	self.foundSplitShape = nil;
	
	self.cutNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.cutNode#node"));
	self.maxCutRadius = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.cutNode#maxRadius"), 1);
	self.maxCutAngle = math.cos(Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.cutNode#maxCutAngle"), 70));
	self.cutSizeX = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.cutNode#sizeX"), 0.15);
	self.cutSizeY = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.cutNode#sizeY"), 1);
	self.cutSizeZ = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.cutNode#sizeZ"), 1);
	
	assert(self.cutNode ~= nil, "TreeShear - Error: cutNode.node is invalid, this play an vital part of the script.");
	
	self.cutState = TreeShear.STATE_READY;
	self.cutAnimation = {};
	self.cutAnimation.name = getXMLString(xmlFile, "vehicle.cutAnimation#name");
	self.cutAnimation.speedScale = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.cutAnimation#speedScale"), 1);
	self.cutAnimation.cutTime = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.cutAnimation#cutTime"), 1);
	
	assert(self.animations[self.cutAnimation.name] ~= nil, "TreeShear - Error: cutAnimation.name is invalid, this play an vital part of the script.");
	
	self.shearState = false;
	self.shearAnimation = {};
	self.shearAnimation.name = getXMLString(xmlFile, "vehicle.shearAnimation#name");
	self.shearAnimation.speedScale = Utils.getNoNil(getXMLFloat(xmlFile, "vehicle.shearAnimation#speedScale"), 1);
	
	self.warnInvalidTree = false;
	self.warnInvalidTreeRadius = false;
end;

function TreeShear:delete()
end;

function TreeShear:readStream(streamId, connection)
	self:setShearState(streamReadBool(streamId), true);
end;

function TreeShear:writeStream(streamId, connection)
	streamWriteBool(streamId, self.shearState);
end;

function TreeShear:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not resetVehicles then
		local animationState = getXMLBool(xmlFile, key .. "#shearState");
		
		if animationState ~= nil then
			self:setShearState(animationState, true);
			
			if animationState ~= nil and self.playAnimation ~= nil then
				AnimatedVehicle.updateAnimationByName(self, self.doorAnimation, 99999999);
			end;
		end;
	end;
	
	return BaseMission.VEHICLE_LOAD_OK;
end;

function TreeShear:getSaveAttributesAndNodes(nodeIdent)
	local attributes = 'shearState="' .. tostring(self.shearState) ..'"';
	
	return attributes;
end;

function TreeShear:mouseEvent(posX, posY, isDown, isUp, button)
end;

function TreeShear:keyEvent(unicode, sym, modifier, isDown)
end;

function TreeShear:update(dt)
	if self:getIsActive() then
		-- Verify that the split shapes still exist (possible that someone has cut them)
		if self.isServer then
			local lostShape = false;
			
			if self.foundSplitShape ~= nil then
                if not entityExists(self.foundSplitShape) then
                    self.foundSplitShape = nil;
                    lostShape = true;
                end;
            end;
			
			if lostShape and self.cutState ~= TreeShear.STATE_READY then
				self:setHeadState(TreeShear.CHANGE_RETURN);
			end;
		end;
		
		if self:getIsActiveForInput(true) then
			if self.foundSplitShape ~= nil then
				if self.cutState == TreeShear.STATE_READY then
					if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
						self:setHeadState(TreeShear.CHANGE_CUT);
					end;
				end;
			end;
		end;
		
		if self.isServer then
			if self.cutState == TreeShear.STATE_CUTTING_TREE then
				if self.foundSplitShape ~= nil then
					if self:getAnimationTime(self.cutAnimation.name) >= self.cutAnimation.cutTime then
						self:setHeadState(TreeShear.CHANGE_FALLING);
						
						local x, y, z = getWorldTranslation(self.cutNode);
						local nx, ny, nz = localDirectionToWorld(self.cutNode, 1, 0, 0);
						local yx, yy, yz = localDirectionToWorld(self.cutNode, 0, 1, 0);

						local currentSplitShape = self.foundSplitShape;
						self.foundSplitShape = nil;
						
						splitShape(currentSplitShape, x, y, z, nx, ny, nz, yx, yy, yz, self.cutSizeY, self.cutSizeZ);
						
						if self.foundSplitShape ~= nil then
							local xD,yD, zD = getWorldTranslation(self.cutNode);
							local nxD, nyD, nzD = localDirectionToWorld(self.cutNode, 1, 0, 0);
							local yxD, yyD, yzD = localDirectionToWorld(self.cutNode, 0, 1, 0);
							local vx, vy, vz = x- xD, y - yD, z - zD;
							local sizeX = Utils.vector3Length(vx, vy, vz);
							removeSplitShapeAttachments(self.foundSplitShape, xD + vx * 3, yD + vy * 3, zD + vz * 3, nxD, nyD, nzD, yxD, yyD, yzD, sizeX * 3 + self.cutSizeX, self.cutSizeY, self.cutSizeZ);
						end;
					end;
				end;
				
			elseif self.cutState == TreeShear.STATE_CUTTING_TREE_IS_FALLING then
				if self:getAnimationTime(self.cutAnimation.name) == 1 then
					self:setHeadState(TreeShear.CHANGE_RETURN);
				end;
				
			elseif self.cutState == TreeShear.STATE_RETURN then
				if self:getAnimationTime(self.cutAnimation.name) == 0 then
					self:setHeadState(TreeShear.CHANGE_RESET);
				end;
			end;
		end;
	end;
end;

function TreeShear:updateTick(dt)
	if self:getIsActive() then
		self.warnInvalidTree = false;
        self.warnInvalidTreeRadius = false;
	
		local x, y, z = getWorldTranslation(self.cutNode);
		local nx, ny, nz = localDirectionToWorld(self.cutNode, 1, 0, 0);
		local yx, yy, yz = localDirectionToWorld(self.cutNode, 0, 1, 0);
		
		if self.lookForSplitShape and self.foundSplitShape == nil then
			local shape, minY, maxY, minZ, maxZ = findSplitShape(x, y, z, nx, ny, nz, yx, yy, yz, self.cutSizeY, self.cutSizeZ);
			
			if shape ~= 0 then
				local splitType = SplitUtil.splitTypes[getSplitType(shape)];
				
				if splitType == nil then -- or not splitType.allowsWoodHarvester then -- we arent the normal cutting head..
					self.warnInvalidTree = true;
				else
					local treeDx, treeDy, treeDz = localDirectionToWorld(shape, 0, 1, 0);
					local cosTreeAngle = Utils.dotProduct(nx, ny, nz, treeDx, treeDy, treeDz);
					
					if (cosTreeAngle >= self.maxCutAngle or cosTreeAngle <= -self.maxCutAngle) then
						local radius = math.max(maxY - minY, maxZ - minZ) * 0.5 * cosTreeAngle;
						
						if radius > self.maxCutRadius then
							self.warnInvalidTreeRadius = true;
						else
							self.foundSplitShape = shape;
						end;
					end;
				end;
			end;
		end;
		
		if self.foundSplitShape ~= nil then
			local minY, maxY, minZ, maxZ = testSplitShape(self.foundSplitShape, x, y, z, nx, ny, nz, yx, yy, yz, self.cutSizeY, self.cutSizeZ)
			
			if (minY == nil or not self.lookForSplitShape) then
				self.foundSplitShape = nil;
			else
				local cutTooLow = false;
				local x, y, z = localToLocal(self.cutNode, self.foundSplitShape, 0, minY, minZ);
				cutTooLow = cutTooLow or y < 0.01;
				local x, y, z = localToLocal(self.cutNode, self.foundSplitShape, 0, minY, maxZ);
				cutTooLow = cutTooLow or y < 0.01;
				local x, y, z = localToLocal(self.cutNode, self.foundSplitShape, 0, maxY, minZ);
				cutTooLow = cutTooLow or y < 0.01;
				local x, y, z = localToLocal(self.cutNode, self.foundSplitShape, 0, maxY, maxZ);
				cutTooLow = cutTooLow or y < 0.01;
				
				if cutTooLow then
					self.foundSplitShape = nil;
				end;
			end;
		end;
	end;
end;

function TreeShear:draw()
    if self:getIsActiveForInput(true) then
		if self.foundSplitShape ~= nil then
			-- if self.cutState == TreeShear.STATE_READY then
				g_currentMission:addHelpButtonText(g_i18n:getText("woodHarvester_cut"), InputBinding.IMPLEMENT_EXTRA2);
			-- else
				-- cutting..
			-- end;
		end;
		
		if self.warnInvalidTreeRadius then
			g_currentMission:showBlinkingWarning(g_i18n:getText("InvalidTreeDiameterWarning"), 1000);
		elseif self.warnInvalidTree then
			g_currentMission:showBlinkingWarning(g_i18n:getText("InvalidTreeTypeWarning"), 1000);
		end;
    end
end;

function TreeShear:attachHydraulicCoupling(vehicle, jointDescIndex, noEventSend)
	if self.hydraulicCouplingIsAttached["hydraulic"] then
		self.lookForSplitShape = true;
	end;
end;

function TreeShear:detachHydraulicCoupling()
	self.lookForSplitShape = false;
end;

function TreeShear:setShearState(animationState, noEventSend)
	self.shearState = animationState;
	
	if self.shearAnimation.name ~= nil then
		if animationState then
			self:playAnimation(self.shearAnimation.name, self.shearAnimation.speedScale, nil, true);
		else
			self:playAnimation(self.shearAnimation.name, -self.shearAnimation.speedScale, nil, true);
		end;
	end;
	
	setShearStateEvent.sendEvent(self, animationState, noEventSend);
end;

function TreeShear:setHeadState(state, noEventSend)
	if state == TreeShear.CHANGE_CUT then
		if self.foundSplitShape ~= nil then
			self:setAnimationTime(self.cutAnimation.name, 0, true);
			self:playAnimation(self.cutAnimation.name, self.cutAnimation.speedScale, self:getAnimationTime(self.cutAnimation.name), true);
			
			self.cutState = TreeShear.STATE_CUTTING_TREE;
		end;
		
	elseif state == TreeShear.CHANGE_FALLING then
		self.cutState = TreeShear.STATE_CUTTING_TREE_IS_FALLING;
		
	elseif state == TreeShear.CHANGE_RETURN then
		self:playAnimation(self.cutAnimation.name, -self.cutAnimation.speedScale, nil, true);
        self.cutState = TreeShear.STATE_RETURN;
	
	elseif state == TreeShear.CHANGE_RESET then
        self.cutState = TreeShear.STATE_READY;
	end;

	setHeadStateEvent.sendEvent(self, state, noEventSend);
end;


-- Event
setShearStateEvent = {};
setShearStateEvent_mt = Class(setShearStateEvent, Event);

InitEventClass(setShearStateEvent, "setShearStateEvent");

function setShearStateEvent:emptyNew()
    local self = Event:new(setShearStateEvent_mt);
    self.className = "setShearStateEvent";
	
    return self;
end;

function setShearStateEvent:new(object, animationState)
    local self = setShearStateEvent:emptyNew()
    self.object = object;
	self.animationState = animationState;
	
    return self;
end;

function setShearStateEvent:readStream(streamId, connection)
    local id = streamReadInt32(streamId);
    self.object = networkGetObject(id);
	self.animationState = streamReadBool(streamId);
	
    self:run(connection);
end;

function setShearStateEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId, self.animationState);	
end;

function setShearStateEvent:run(connection)
	self.object:setShearState(self.animationState, true);
	
	if not connection:getIsServer() then
		g_server:broadcastEvent(setShearStateEvent:new(self.object, self.animationState), nil, connection, self.object);
	end;
end;

function setShearStateEvent.sendEvent(object, animationState, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(setShearStateEvent:new(object, animationState), nil, nil, object);
		else
			g_client:getServerConnection():sendEvent(setShearStateEvent:new(object, animationState));
		end;
	end;
end;


-- Event
setHeadStateEvent = {};
setHeadStateEvent_mt = Class(setHeadStateEvent, Event);

InitEventClass(setHeadStateEvent, "setHeadStateEvent");

function setHeadStateEvent:emptyNew()
    local self = Event:new(setHeadStateEvent_mt);
    self.className = "setHeadStateEvent";
	
    return self;
end;

function setHeadStateEvent:new(object, stateId)
    local self = setHeadStateEvent:emptyNew()
    self.object = object;
	self.stateId = stateId;
	
    return self;
end;

function setHeadStateEvent:readStream(streamId, connection)
    local id = streamReadInt32(streamId);
    self.object = networkGetObject(id);
	self.stateId = streamReadInt8(streamId);
	
    self:run(connection);
end;

function setHeadStateEvent:writeStream(streamId, connection)
    streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteInt8(streamId, self.stateId);	
end;

function setHeadStateEvent:run(connection)
	self.object:setHeadState(self.stateId, true);
	
	if not connection:getIsServer() then
		g_server:broadcastEvent(setHeadStateEvent:new(self.object, self.stateId), nil, connection, self.object);
	end;
end;

function setHeadStateEvent.sendEvent(object, stateId, noEventSend)
	if noEventSend == nil or noEventSend == false then
		if g_server ~= nil then
			g_server:broadcastEvent(setHeadStateEvent:new(object, stateId), nil, nil, object);
		else
			g_client:getServerConnection():sendEvent(setHeadStateEvent:new(object, stateId));
		end;
	end;
end;